﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections.Specialized;
using System.ComponentModel;
using System.Threading;
using System.Collections;
using Windows.UI.Xaml.Data;
using System.Diagnostics;
using Windows.Foundation;
using System.IO;

namespace DataVirtualization
{
    /// <summary>
    /// Derived VirtualizatingCollection, performing loading asychronously.
    /// </summary>
    /// <typeparam name="T">The type of items in the collection</typeparam>
    public class AsyncVirtualizingCollection<T> :
        VirtualizingCollection<T>,
        INotifyCollectionChanged, 
        INotifyPropertyChanged
      , ISupportIncrementalLoading
        where T : class
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="AsyncVirtualizingCollection&lt;T&gt;"/> class.
        /// </summary>
        /// <param name="itemsProvider">The items provider.</param>
        /// <param name="pageSize">Size of the page.</param>
        /// <param name="pageTimeout">The page timeout.</param>
        public AsyncVirtualizingCollection(IItemsProvider<T> itemsProvider, int pageSize, int pageTimeout)
            : base(itemsProvider, pageSize, pageTimeout)
        {
        }

        #region SynchronizationContext
        #endregion

        #region INotifyCollectionChanged

        /// <summary>
        /// Occurs when the collection changes.
        /// </summary>
        public event NotifyCollectionChangedEventHandler CollectionChanged;

        /// <summary>
        /// Raises the <see cref="E:CollectionChanged"/> event.
        /// </summary>
        /// <param name="e">The <see cref="System.Collections.Specialized.NotifyCollectionChangedEventArgs"/> instance containing the event data.</param>
        protected virtual void OnCollectionChanged(NotifyCollectionChangedEventArgs e)
        {
            NotifyCollectionChangedEventHandler h = CollectionChanged;
            if (h != null)
                h(this, e);
        }

        /// <summary>
        /// Fires the collection reset event.
        /// </summary>
        private void FireCollectionReset()
        {
            NotifyCollectionChangedEventArgs e = new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Reset);
            OnCollectionChanged(e);
        }

        #endregion

        #region INotifyPropertyChanged

        /// <summary>
        /// Occurs when a property value changes.
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged;

        /// <summary>
        /// Raises the <see cref="E:PropertyChanged"/> event.
        /// </summary>
        /// <param name="e">The <see cref="System.ComponentModel.PropertyChangedEventArgs"/> instance containing the event data.</param>
        protected virtual void OnPropertyChanged(PropertyChangedEventArgs e)
        {
            PropertyChangedEventHandler h = PropertyChanged;
            if (h != null)
                h(this, e);
        }

        /// <summary>
        /// Fires the property changed event.
        /// </summary>
        /// <param name="propertyName">Name of the property.</param>
        private void FirePropertyChanged(string propertyName)
        {
            PropertyChangedEventArgs e = new PropertyChangedEventArgs(propertyName);
            OnPropertyChanged(e);
        }

        #endregion

        #region IsLoading

        private bool _isLoading;

        /// <summary>
        /// Gets or sets a value indicating whether the collection is loading.
        /// </summary>
        /// <value>
        /// 	<c>true</c> if this collection is loading; otherwise, <c>false</c>.
        /// </value>
        public bool IsLoading
        {
            get
            {
                return _isLoading;
            }
            set
            {
                if (value != _isLoading)
                {
                    _isLoading = value;
                    FirePropertyChanged("IsLoading");
                }
            }
        }

        private bool _isInitializing;

        public bool IsInitializing
        {
            get
            {
                return _isInitializing;
            }
            set
            {
                if (value != _isInitializing)
                {
                    _isInitializing = value;
                    FirePropertyChanged("IsInitializing");
                }
            }
        }
        #endregion

        #region Load overrides

        internal void ForceLoadCount()
        {
            LoadCount();
        }

        /// <summary>
        /// Asynchronously loads the count of items.
        /// </summary>
        protected override void LoadCount()
        {
            if (Count == 0)
                IsInitializing = true;

            LoadCountAsync();
        }

        /// <summary>
        /// Performed on background thread.
        /// </summary>
        /// <param name="args">None required.</param>
        private async void LoadCountAsync()
        {
            int count = await FetchCount();

            //update count and reset collection only in case of new count differs from what we have now
            if(this.Count != count)
                this.TakeNewCount(count);
            IsInitializing = false;
        }

        private void TakeNewCount(int newCount)
        {
            if (newCount != this.Count)
            {
                this.Count = newCount;
                this.EmptyCache();
                FireCollectionReset();
            }
        }

        /// <summary>
        /// Asynchronously loads the page.
        /// </summary>
        /// <param name="index">The index.</param>
        protected override void LoadPage(int pageIndex, int pageLength)
        {
            IsLoading = true;
            //ThreadPool.QueueUserWorkItem(LoadPageWork, new int[] { pageIndex, pageLength });
            LoadPageAsync(pageIndex, pageLength);
        }

        /// <summary>
        /// Performed on background thread.
        /// </summary>
        /// <param name="args">Index of the page to load.</param>
        private async void LoadPageAsync(int pageIndex, int pageLength)
        {
            int overallCount = 0;
            IList<T> dataItems = await FetchPage(pageIndex, pageLength);//, out overallCount);
            overallCount = await ItemsProvider.FetchCount(); //dataItems.Count;

            this.TakeNewCount(overallCount);
            PopulatePage(pageIndex, dataItems);
            IsLoading = false;
        }

        #endregion

        private bool _hasMoreItems = true;
        /// <summary>
        /// Gets/sets whether the ListViewBase (e.g. GridView/ListView) should continue loading data from
        /// source
        /// </summary>
        public bool HasMoreItems
        {
            get
            {
                CheckMoreItemsAsync();
                return _hasMoreItems;
            }
            set
            {
                _hasMoreItems = value;
                OnPropertyChanged(new PropertyChangedEventArgs("HasMoreItems"));
            }
        }

        private async void CheckMoreItemsAsync()
        {
            int updatedCount = await ItemsProvider.FetchCount();
            bool result = updatedCount > this.Count;
            HasMoreItems = result;
        }

        public IAsyncOperation<LoadMoreItemsResult> LoadMoreItemsAsync(uint count)
        {
            return new IncrementalLoader<T>(this, count);
        }
    }



    /// <summary>
    /// This class is responsible for retrieving data
    /// </summary>
    /// <typeparam name="T"></typeparam>
    public class IncrementalLoader<T> : IAsyncOperation<LoadMoreItemsResult>
        where T: class
    {
        private AsyncStatus _asyncStatus = AsyncStatus.Started;
        private LoadMoreItemsResult _results;

        public IncrementalLoader(AsyncVirtualizingCollection<T> collection, uint count)
        {
            PullDataIncrementalData(collection, count);
        }

        public async void PullDataIncrementalData(AsyncVirtualizingCollection<T> collection, uint count)
        {
            try
            {
                collection.IsLoading = true;
                int nextPageIndex = collection.LastPageIndex + 1;

                collection.ForceLoadCount();

                //IEnumerable<T> newItems = await incrementalLoadingCollection.PullDataAsync(count);

                //if (newItems != null)
                //{
                //    if (newItems.Count() > 0)
                //    {
                //        foreach (T item in newItems)
                //        {
                //            incrementalLoadingCollection.Add(item);
                //        }
                //    }
                //    else
                //    {
                //        incrementalLoadingCollection.HasMoreItems = false;
                //    }
                //}
                //else
                //{
                //    throw new InvalidDataException();
                //}

                // On success, increment page
                _asyncStatus = AsyncStatus.Completed;
                //incrementalLoadingCollection.CurrentPage++;
            }
            catch
            {
                _results.Count = 0;
                _asyncStatus = AsyncStatus.Error;
                collection.HasMoreItems = false;
            }

            collection.IsLoading = false;
            if (Completed != null)
            {
                Completed(this, _asyncStatus);
            }
        }

        public AsyncOperationCompletedHandler<LoadMoreItemsResult> Completed { get; set; }

        public LoadMoreItemsResult GetResults()
        {
            return _results;
        }

        public void Cancel()
        {
            throw new NotImplementedException();
        }

        public void Close()
        {
        }

        public Exception ErrorCode
        {
            get { throw new NotImplementedException(); }
        }

        public uint Id
        {
            get { throw new NotImplementedException(); }
        }

        public AsyncStatus Status
        {
            get { return _asyncStatus; }
        }
    }
}
