//===================================================================================
// Microsoft patterns & practices
// Composite Application Guidance for Windows Presentation Foundation and Silverlight
//===================================================================================
// Copyright (c) Microsoft Corporation.  All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
// FITNESS FOR A PARTICULAR PURPOSE.
//===================================================================================
// The example companies, organizations, products, domain names,
// e-mail addresses, logos, people, places, and events depicted
// herein are fictitious.  No association with any real company,
// organization, product, domain name, email address, logo, person,
// places, or events is intended or should be inferred.
//===================================================================================
using System;
using System.Architecture;
using System.Collections.Generic;
using System.Composition;
using System.Events;
using System.Globalization;
using System.Reflection;
using System.Linq;
//using Microsoft.Practices.Prism.Events;
//using Microsoft.Practices.ServiceLocation;
//using Prism.Desktop.Metro;

namespace Microsoft.Practices.Prism.Regions
{
    /// <summary>
    /// Defines a registry for the content of the regions used on View Discovery composition.
    /// </summary>
    [Shared]
    [Export(typeof(IRegionViewRegistry))]
    public class RegionViewRegistry : IRegionViewRegistry
    {
        //private readonly IServiceLocator locator;
        private readonly ListDictionary<string, Func<object>> registeredContent = new ListDictionary<string, Func<object>>();
        private IMVVMContentLoader MVVMLoader;
        //private readonly WeakDelegatesManager contentRegisteredListeners = new WeakDelegatesManager();

        IEventPublisher EventAggregator;

        /// <summary>
        /// Creates a new instance of the <see cref="RegionViewRegistry"/> class.
        /// </summary>
        /// <param name="locator"><see cref="IServiceLocator"/> used to create the instance of the views from its <see cref="Type"/>.</param>
        [ImportingConstructor]
        public RegionViewRegistry(/*IServiceLocator locator*/
            [Import] 
            IEventPublisher eventAggregator,
            [Import]
            IMVVMContentLoader mvvmLoader)
        {
            //this.locator = locator;
            EventAggregator = eventAggregator;
            MVVMLoader = mvvmLoader;

            if (MVVMLoader != null)
            {
                var factories = MVVMLoader.GetViewFactories(ef => ef.Metadata.DefaultRegions != null && ef.Metadata.DefaultRegions.Count() > 0);
                var result = from f in factories
                             let m = f.Metadata
                             from r in m.DefaultRegions
                             select new 
                             { 
                                 RegionName = r,
                                 CreateDelegate = new Func<object>(() => MVVMLoader.GetView(f, true)) 
                             };

                foreach(var f in result)
                    RegisterViewWithRegion(f.RegionName, f.CreateDelegate);

            }
        }

        ///// <summary>
        ///// Occurs whenever a new view is registered.
        ///// </summary>
        //public event EventHandler<ViewRegisteredEventArgs> ContentRegistered
        //{
        //    add { this.contentRegisteredListeners.AddListener(value); }
        //    remove { this.contentRegisteredListeners.RemoveListener(value); }
        //}

        /// <summary>
        /// Returns the contents registered for a region.
        /// </summary>
        /// <param name="regionName">Name of the region which content is being requested.</param>
        /// <returns>Collection of contents registered for the region.</returns>
        public IEnumerable<object> GetContents(string regionName)
        {
            List<object> items = new List<object>();
            foreach (Func<object> getContentDelegate in this.registeredContent[regionName])
            {
                items.Add(getContentDelegate());
            }

            return items;
        }

        /// <summary>
        /// Registers a content type with a region name.
        /// </summary>
        /// <param name="regionName">Region name to which the <paramref name="viewType"/> will be registered.</param>
        /// <param name="viewType">Content type to be registered for the <paramref name="regionName"/>.</param>
        public void RegisterViewWithRegion(string regionName, Type viewType)
        {
            this.RegisterViewWithRegion(regionName, () => this.CreateInstance(viewType));
        }

        /// <summary>
        /// Registers a delegate that can be used to retrieve the content associated with a region name. 
        /// </summary>
        /// <param name="regionName">Region name to which the <paramref name="getContentDelegate"/> will be registered.</param>
        /// <param name="getContentDelegate">Delegate used to retrieve the content associated with the <paramref name="regionName"/>.</param>
        public void RegisterViewWithRegion(string regionName, Func<object> getContentDelegate)
        {
            this.registeredContent.Add(regionName, getContentDelegate);
            this.OnContentRegistered(new ViewRegisteredMessage(regionName, getContentDelegate));
        }

        /// <summary>
        /// Creates an instance of a registered view <see cref="Type"/>. 
        /// </summary>
        /// <param name="type">Type of the registered view.</param>
        /// <returns>Instance of the registered view.</returns>
        protected virtual object CreateInstance(Type type)
        {
            return null;
            /////////////////////////////////////////////////////////return this.locator.GetInstance(type);
        }

        private void OnContentRegistered(ViewRegisteredMessage e)
        {
            try
            {
                EventAggregator.Publish<ViewRegisteredMessage>(e);
                //this.contentRegisteredListeners.Raise(this, e);
            }
            catch (TargetInvocationException ex)
            {
                Exception rootException;
                if (ex.InnerException != null)
                {
                     rootException = ex.InnerException.GetRootException();
                }
                else
                {
                    rootException = ex.GetRootException();
                }

                throw new ViewRegistrationException(string.Format(CultureInfo.CurrentCulture,
                    ResourceHelper.OnViewRegisteredException, e.RegionName, rootException), ex.InnerException);
            }
        }
    }
}